library amf;
import 'dart:html';
import 'dart:typed_data';
import 'dart:async';
import 'dart:mirrors';
//import 'package:logging_handlers/logging_handlers_shared.dart';
part 'amf_archiver.dart';
part 'amf_unarchiver.dart';

part 'externalizable.dart';
part 'flex_data_types.dart';
part 'gds_remote_object.dart';
part 'types.dart';

class AMF {

  static Map _typeMap = new Map();
  static Map _functionMap = new Map();

  static Map<Type, String> _nameMap = new Map();

  static void init() {
    registerAnnotatedClass(FlexAcknowledgeMessage);
    registerAnnotatedClass(FlexArrayCollection);
    registerAnnotatedClass(FlexRemotingMessage);
  }

  static void registerAnnotatedClass(Type t, [Function f]) {
    print("Register type ${t}");
    ClassMirror classMirror = reflectClass(t);
    for (InstanceMirror metadata in classMirror.metadata) {
      if (metadata.reflectee is RemoteClass) {
        print("Registered classmirror for ${metadata.reflectee.name}");
        _typeMap[metadata.reflectee.name] = classMirror;
        _nameMap[t] = metadata.reflectee.name;
        if (f != null) _functionMap[metadata.reflectee.name] = f;
      }
    }
  }

  static ClassMirror findClassMirror(String name) {
    if (_typeMap.containsKey(name)) return _typeMap[name];

    print("Having to look for classmirror");

    for (var lib in currentMirrorSystem().libraries.values) {
      var mirror = lib.declarations[MirrorSystem.getSymbol(name)];
      if (mirror != null) return mirror;
    }
    //throw new ArgumentError("Class $name does not exist");
    return null;
  }

  static Object createClassByName(String classname) {
    print("CreateclassByNAme ${classname}");
    if (_functionMap[classname] != null) return _functionMap[classname]();
    ClassMirror classMirror = findClassMirror(classname);
    if (classMirror == null) throw new ArgumentError(
        "Class $classname does not exist");

    return classMirror.newInstance(const Symbol(''), []).reflectee;

  }

  static String classNameForClass(type) {
    String name = _nameMap[type];
    if (name == null) {
      throw new StateError("No such class for ${type}");
    }
    return name;
  }
}

class Double {

  double num;
  Double(this.num);
}

class XML {
  Object xml;
  XML(this.xml);
}


class AMF3TraitsInfo {
  String _className;
  bool _dynamic = false;
  bool _externalizable = false;
  int _count;
  var _properties = [];

  void addProperty(String prop) {
    _properties.add(prop);
  }
}

class AmfRemotingCall {


  String serviceUrl;
  String service;
  String method;
  var arguments;
  var amfVersion;
  Map amfHeaders;

  AmfRemotingCall() {
    /*  m_receivedData = nil;
    m_connection = nil;
    m_delegate = nil;
    m_isLoading = NO;
    m_amfVersion = kAMF3Version;
    m_error = nil;
    m_amfHeaders = nil;
    
    m_request = [[NSMutableURLRequest alloc] init];
    [m_request setHTTPMethod:@"POST"];
    [m_request setValue:@"application/x-amf" forHTTPHeaderField:@"Content-Type"];
    [m_request setValue:@"CocoaAMF" forHTTPHeaderField:@"User-Agent"];
   */

    amfVersion = 3;

  }

  void start(Function f) {
    /* 
    if (m_isLoading)
    {
      return;
    }

    AMFActionMessage *message = [[AMFActionMessage alloc] init];
    message.version = m_amfVersion;
    NSString *targetURI = @"null";
    if (m_service != nil && m_method != nil)
      targetURI = [NSString stringWithFormat:@"%@.%@", m_service, m_method];
    [message addBodyWithTargetURI:targetURI responseURI:[self _nextResponseURI] data:m_arguments];
    if (m_amfHeaders != nil)
      message.headers = [m_amfHeaders allValues];
    
    [m_request setHTTPBody:[message data]];
    [message release];
    
    m_error = nil;
    m_receivedData = [[NSMutableData alloc] init];
    m_connection = [[NSURLConnection alloc] initWithRequest:m_request delegate:self];
    
    m_isLoading = YES;
    */

    AMFActionMessage message = new AMFActionMessage();
    message.version = amfVersion;

    // TODO: Generate new URI value
    String nextResponseUri = "/1";
    message.addBody("${this.service}.${this.method}", nextResponseUri,
        this.arguments);

    if (amfHeaders != null) message.headers = amfHeaders.values;

    HttpRequest request = new HttpRequest();
    //    [m_request setValue:@"application/x-amf" forHTTPHeaderField:@"Content-Type"];
    //[m_request setValue:@"CocoaAMF" forHTTPHeaderField:@"User-Agent"];



    request.open('POST', serviceUrl);
    request.responseType = "arraybuffer";
    request.setRequestHeader("Content-Type", "application/x-amf");
    //request.setRequestHeader("User-Agent", "DartAMF");

    Object data = message.data;

    HttpRequest.request(serviceUrl, method: "POST", responseType: "arraybuffer",
        sendData: data).then((request) {
      Uint8List buffer = new Uint8List.view(request.response, 0,
          (request.response as ByteBuffer).lengthInBytes);
      // this is a hack for dart2js because of a bug
      print(buffer);
    });



    /*
    request.on.readyStateChange.add((Event e) {
      print("hello ${e.type} ${request.readyState}");
      f(this);
    });
    
    */
  }
}



class AMFMessageHeader {
  String name;
  bool mustUnderstand;
  var data;
}

class AMFMessageBody {
  String targetURI;
  String responseURI;
  var data;
}

class AMFActionMessage {
  List<AMFMessageHeader> headers = new List<AMFMessageHeader>();
  List<AMFMessageBody> bodies = new List<AMFMessageBody>();
  int version = 3;

  AMFActionMessage() {}

  AMFActionMessage.initWithData(Uint8List data) {
    var bb = new ByteData.view(data.buffer);
    AMFUnarchiver ba = new AMFUnarchiver(AMFVersion.kAMF0Version, bb);

    version = ba.decodeUnsignedShort();
    int numHeaders = ba.decodeUnsignedShort();

    for (int i = 0; i < numHeaders; i++) {
      AMFMessageHeader header = new AMFMessageHeader();
      header.name = ba.decodeUTF();
      header.mustUnderstand = ba.decodeBool();
      ba.decodeUnsignedInt();
      header.data = ba.decodeObject();
      headers.add(header);
    }

    int numBodies = ba.decodeUnsignedShort();

    for (int i = 0; i < numBodies; i++) {
      AMFMessageBody body = new AMFMessageBody();
      body.targetURI = ba.decodeUTF();
      body.responseURI = ba.decodeUTF();
      ba.decodeUnsignedInt();
      body.data = ba.decodeObject();
      bodies.add(body);
    }

  }

  void addBody(String targetURI, String responseURI, var data) {
    AMFMessageBody body = new AMFMessageBody();
    body.targetURI = targetURI;
    body.responseURI = responseURI;
    body.data = data;
    bodies.add(body);
  }

  Object get data {

    AMFArchiver ba = new AMFArchiver(0);


    ba.encodeUnsignedShort(this.version);
    ba.encodeUnsignedShort(this.headers.length);


    for (AMFMessageHeader header in headers) {


      ba.encodeUTF(header.name);
      ba.encodeBool(header.mustUnderstand);


      AMFArchiver headerBa = new AMFArchiver(version);
      if (version == 3) {
        headerBa.encodeUnsignedChar(AMF0Type.kAMF0AVMPlusObjectType);
      }


      headerBa.encodeObject(header.data);
      ba.encodeUnsignedInt(headerBa.data.length);
      ba.encodeDataObject(headerBa.data);

    }
    //  [ba encodeUnsignedShort:[m_bodies count]];
    //
    ba.encodeUnsignedShort(this.bodies.length);


    for (AMFMessageBody body in bodies) {
      if (body.targetURI == null) ba.encodeUTF("null"); else ba.encodeUTF(
          "${body.targetURI}");

      if (body.responseURI == null) ba.encodeUTF("null"); else ba.encodeUTF(
          "${body.responseURI}");


      AMFArchiver bodyBa = new AMFArchiver(version);

      if (version == 3) {
        bodyBa.encodeUnsignedChar(AMF0Type.kAMF0AVMPlusObjectType);
      }

      bodyBa.encodeObject(body.data);
      // Think that this should be a list or an array, but it isn't.



      ba.encodeUnsignedInt(bodyBa.data.length);
      ba.encodeDataObject(bodyBa.data);

    }
    //  NSData *data = [[ba.data retain] autorelease];
    //  [ba release];
    //
    return ba.data;

  }
}



class ASObject {
  String _type;
  bool _isExternalizable = false;
  Map _properties = new Map();
  List _data = null;

  void addObject(obj) {
    if (_data == null) _data = [];

    _data.add(obj);
  }

  void setValue(String key, value) {
    _properties[key] = value;
  }

  int get count {
    if (_isExternalizable) return _data.length;
    return _properties.length;
  }

  void operator []=(key, value) {
    _properties[key] = value;
  }

  Object operator [](key) {
    return _properties[key];
  }

  noSuchMethod(Invocation invocation) {
    // TODO: is this *really* the only way?
    String name = invocation.memberName.toString();
    name = name.substring(8, name.length - 2);
   
    return _properties[name];
  }
}

class AMFPlainData {

  var _data;

  AMFPlainData.plainDataWithData(data) {
    _data = data;
  }

}
