part of amf;


class AMFArchiver {
  Uint8List _data;
  var _buffer;
  ByteData m_data;
  int m_position = 0;

  //NSMutableArray *m_objectTable;
  //NSMutableArray *m_serializationStack;
  //NSMutableArray *m_writeStack;
  //NSMutableDictionary *m_registeredClasses;
  //ASObject *m_currentObjectToSerialize;
  //NSObject *m_currentObjectToWrite;

  ASObject _currentObjectToSerialize;
  List _writeStack;
  var _currentObjectToWrite;
  List _serializationStack;

  factory AMFArchiver(int amfVersion) {
    if (amfVersion == 3) return new AMF3Archiver.create(); else return
        new AMF0Archiver.create();
  }

  AMFArchiver.create() {

    _data = new Uint8List(102400);
    _buffer = _data.buffer;
    m_data = new ByteData.view(_buffer);
  }

  void encodeUnsignedShort(int value) {
    _ensureLength(2);
    m_data.setUint16(m_position, value);
    m_position += 2;
  }

  void encodeUTF(String s) {
    encodeUnsignedShort(s.length);

    _ensureLength(s.length);
    for (int i = 0; i < s.length; i++) {
      var cc = s.codeUnitAt(i);
      m_data.setUint8(m_position++, cc);
    }
  }

  void encodeBool(bool value) {
    encodeUnsignedChar: (value ? 1 : 0);
  }

  void encodeUnsignedChar(value) {
    if (_currentObjectToSerialize != null) {
      _currentObjectToSerialize.addObject(value);
      _ensureIntegrityOfSerializedObject();
      return;
    }
    _ensureLength(1);
    m_data.setUint8(m_position++, value);
  }

  void encodeUnsignedInt29(uint32value) {
    if (_currentObjectToSerialize != null) {
      _currentObjectToSerialize.addObject(uint32value);
      _ensureIntegrityOfSerializedObject();
      return;
    }
    if (uint32value < 0x80) {
      _ensureLength(1);
      m_data.setUint8(m_position++, uint32value);
    } else if (uint32value < 0x4000) {
      _ensureLength(2);
      m_data.setUint8(m_position++, ((uint32value >> 7) & 0x7F) | 0x80);
      m_data.setUint8(m_position++, (uint32value & 0x7F));
    } else if (uint32value < 0x200000) {
      _ensureLength(3);
      m_data.setUint8(m_position++, ((uint32value >> 14) & 0x7F) | 0x80);
      m_data.setUint8(m_position++, ((uint32value >> 7) & 0x7F) | 0x80);
      m_data.setUint8(m_position++, (uint32value & 0x7F));
    } else {
      _ensureLength(4);
      m_data.setUint8(m_position++, ((uint32value >> 22) & 0x7F) | 0x80);
      m_data.setUint8(m_position++, ((uint32value >> 15) & 0x7F) | 0x80);
      m_data.setUint8(m_position++, ((uint32value >> 8) & 0x7F) | 0x80);
      m_data.setUint8(m_position++, (uint32value & 0x7F));
    }
  }

  void encodeObject(o) {
    //print("Encode " + );

    if (o is String) {
      if (_currentObjectToSerialize != null) {
        _currentObjectToSerialize.addObject(o);
        _ensureIntegrityOfSerializedObject();
        return;
      }
      _encodeString(o, false);
    } else if (o is List) {
      // TODO: g_options & AMFArchiverPackArrayOption)
      if (_currentObjectToSerialize != null) {
        _currentObjectToSerialize.addObject(o);
        _ensureIntegrityOfSerializedObject();
        return;
      }
      _setCollectionWriteContext(o);
      _encodeArray(o);
      _restoreCollectionWriteContext();
    } else if (o is ASObject) {
      if (_currentObjectToSerialize != null) {
        _currentObjectToSerialize.addObject(o);
        _ensureIntegrityOfSerializedObject();
        return;
      }
      _setCollectionWriteContext(o);
      _encodeASObject(o);
      _restoreCollectionWriteContext();
    } else if (o is int) {
      _encodeInteger(o);
    } else if (o is Double) {
      _encodeDouble(o.num);
    } else if (o is double) {
      _encodeDouble(o);
    } else if( o is bool)
    {
      _encodeBool(o);
    }
    else if (o is XML) {
      _encodeXML(o.xml);
    } else if (o == null) {
      if (_currentObjectToSerialize != null) {
        _currentObjectToSerialize.addObject(null);
        _ensureIntegrityOfSerializedObject();
        return;
      }
      _encodeNull();
    } else {
      if (_currentObjectToSerialize != null) {
        _currentObjectToSerialize.addObject(o);
        _ensureIntegrityOfSerializedObject();
        return;
      }
      _encodeCustomObject(o);
    }


  }

  void encodeObjectKeyValue(String key, value) {
    _currentObjectToSerialize.setValue(key, value);
    _ensureIntegrityOfSerializedObject();
  }

  void encodeDoubleKeyValue(String key, double value) {
    // TODO : JS / Dart diff
    _currentObjectToSerialize.setValue(key, new Double(value));
    _ensureIntegrityOfSerializedObject();
  }

  void encodeXMLKeyValue(String key, value) {
    _currentObjectToSerialize.setValue(key, new XML(value));
    _ensureIntegrityOfSerializedObject();
  }

  void _encodeASObject(ASObject value) {
    throw new StateError("ERROR: Virtual");
  }

  void _encodeInteger(int value) {
    encodeUnsignedChar(AMF0Type.kAMF0NumberType);
    encodeDouble(value.toDouble());
  }

  void encodeUnsignedInt(ui) {
    _ensureLength(4);
    m_data.setUint32(m_position, ui);
    m_position += 4;
  }

  void _encodeDouble(double value) {
    encodeUnsignedChar(AMF0Type.kAMF0NumberType);
    encodeDouble(value);
  }
  
  void _encodeBool(bool value) {
    encodeUnsignedChar(AMF0Type.kAMF0BooleanType);
    encodeBool(value);
  }

  void encodeDouble(double value) {
    _ensureLength(8);
    m_data.setFloat64(m_position, value);
    m_position += 8;
  }

  void encodeDataObject(value) {
    if (_currentObjectToSerialize != null) {
      _currentObjectToSerialize.addObject(new AMFPlainData.plainDataWithData(
          value));
      _ensureIntegrityOfSerializedObject();
      return;
    }

    _data.setRange(m_position, m_position + value.length, value);
    m_position += value.length;


    /* [m_data appendData:value];
    _data.setRange(m_position, , end, iterable, skipCount)
    
    
    m_bytes = [m_data mutableBytes];
    m_position = [m_data length];

    */
  }

  void _ensureIntegrityOfSerializedObject() {}

  /** ByteDataView **/
  get data {
    return new Uint8List.view(_buffer, 0, m_position);
  }

  void _ensureLength(int length) {}

  void _encodeCustomObject(value) {

    ASObject lastObj = _currentObjectToSerialize;
    ASObject obj = _currentObjectToSerialize = new ASObject();


    if (_serializationStack == null) {
      _serializationStack = [];
    }

    _serializationStack.add(_currentObjectToSerialize);

    obj._type = AMF.classNameForClass(value.runtimeType);
    if (obj._type == null) value.runtimeType.toString();

    value.writeObject(this);

    _currentObjectToSerialize = lastObj;
    _serializationStack.removeLast();


    if (lastObj == null) {
      encodeObject(obj);
    }

  }

  void _encodeNull() {
    encodeUnsignedChar(AMF0Type.kAMF0NullType);
  }

  void _encodeString(String value, bool omitType) {
    List codeUnits = value.codeUnits;

    if (codeUnits.length > 0xffff) {
      if (!omitType) encodeUnsignedChar(AMF0Type.kAMF0LongStringType);
      encodeUnsignedInt(codeUnits.length);
    } else {
      if (!omitType) encodeUnsignedChar(AMF0Type.kAMF0StringType);
      encodeUnsignedShort(codeUnits.length);
    }
    encodeDataObject(codeUnits);
  }

  List _objectTable = [];

  void _encodeArray(List value) {
    if (_objectTable.indexOf(value) >= 0) {
      encodeUnsignedChar(AMF0Type.kAMF0ReferenceType);
      encodeUnsignedShort(_objectTable.indexOf(value));
      return;
    }
    _objectTable.add(value);
    encodeUnsignedChar(AMF0Type.kAMF0StrictArrayType);
    encodeUnsignedInt(value.length);
    value.forEach((e) {
      encodeObject(e);
    });
  }

  void _setCollectionWriteContext(obj) {
    if (_writeStack == null) {
      _writeStack = [];
    }
    _writeStack.add(obj);
    _currentObjectToWrite = obj;
  }
  void _restoreCollectionWriteContext() {
    _writeStack.removeLast();
    if (_writeStack.length > 0) _currentObjectToWrite = _writeStack.last; else
        _currentObjectToWrite = null;
  }
}


class AMF0Archiver extends AMFArchiver {
  AMF0Archiver.create(): super.create() {}
  void encodeUTF(String s) {
    if (s == null) {
      encodeUnsignedShort(0);
      return;
    }
    List codeUnits = s.codeUnits;
    if (codeUnits.length > 0xffff) {
      encodeUnsignedInt(codeUnits.length);
    } else {
      encodeUnsignedShort(codeUnits.length);
    }
    encodeDataObject(codeUnits);
  }

  void _ensureIntegrityOfSerializedObject() {
    throw "Error";
  }

  void _encodeString(String value, bool omitType) {
    throw "Error";
    //void _encodeString:(NSString *)value omitType:(BOOL)omitType
    //{
    //  NSData *stringData = [value dataUsingEncoding:NSUTF8StringEncoding];
    //  if ([stringData length] > 0xFFFF)
    //  {
    //    omitType ?: [self encodeUnsignedChar:kAMF0LongStringType];
    //    [self encodeUnsignedInt:[stringData length]];
    //  }
    //  else
    //  {
    //    omitType ?: [self encodeUnsignedChar:kAMF0StringType];
    //    [self encodeUnsignedShort:[stringData length]];
    //  }
    //  [self encodeDataObject:stringData];
    //}
  }
  void _encodeArray(List value) {
    throw "Error";
    //  if ([m_objectTable indexOfObjectIdenticalTo:value] != NSNotFound)
    //  {
    //    [self encodeUnsignedChar:kAMF0ReferenceType];
    //    [self encodeUnsignedShort:[m_objectTable indexOfObject:value]];
    //    return;
    //  }
    //  [m_objectTable addObject:value];
    //  [self encodeUnsignedChar:kAMF0StrictArrayType];
    //  [self encodeUnsignedInt:[value count]];
    //  for (id obj in value)
    //  {
    //    [self encodeObject:obj];
    //  }
  }

  void _encodeDictionary(value) {
    throw "Error";
    //void _encodeDictionary:(NSDictionary *)value
    //{
    //  if ([m_objectTable indexOfObjectIdenticalTo:value] != NSNotFound)
    //  {
    //    [self encodeUnsignedChar:kAMF0ReferenceType];
    //    [self encodeUnsignedShort:[m_objectTable indexOfObject:value]];
    //    return;
    //  }
    //  [m_objectTable addObject:value];
    //
    //  // empty ecma arrays won't get parsed properly. seems like a bug to me
    //  if ([value count] == 0)
    //  {
    //    // so we write a generic empty object
    //    [self encodeUnsignedChar:kAMF0ObjectType];
    //    [self encodeUnsignedShort:0];
    //    [self encodeUnsignedChar:kAMF0ObjectEndType];
    //    return;
    //  }
    //  [self encodeUnsignedChar:kAMF0ECMAArrayType];
    //  //[self encodeUnsignedInt:[value count]];
    //  // PyAMF does always write 0 instead of length
    //  // @TODO look how flash handles this
    //  [self encodeUnsignedInt:0];
    //  for (NSString *key in value)
    //  {
    //    [self _encodeString:key omitType:YES];
    //    [self encodeObject:[value objectForKey:key]];
    //  }
    //  [self encodeUnsignedShort:0];
    //  [self encodeUnsignedChar:kAMF0ObjectEndType];
  }

  void _encodeASObject(ASObject value) {

    if (_objectTable.indexOf(value) != -1) {
      encodeUnsignedChar(AMF0Type.kAMF0ReferenceType);
      encodeUnsignedShort(_objectTable.indexOf(value));
      return;
    }
    _objectTable.add(value);
    if (value._type == null) {
      encodeUnsignedChar(AMF0Type.kAMF0ObjectType);
      //[self encodeUnsignedShort:0];
    } else {
      encodeUnsignedChar(AMF0Type.kAMF0TypedObjectType);
      _encodeString(value._type, true);
    }
    value._properties.forEach((k, v) {
      _encodeString(k, true);
      encodeObject(v);
    });

    encodeUnsignedShort(0);
    encodeUnsignedChar(AMF0Type.kAMF0ObjectEndType);
  }

  void _encodeNumber(value) {
    throw "Error";
    //  if ([[value className] isEqualToString:@"NSCFBoolean"])
    //  {
    //    [self encodeUnsignedChar:kAMF0BooleanType];
    //    [self encodeBool:[value boolValue]];
    //    return;
    //  }
    //  [self encodeUnsignedChar:kAMF0NumberType];
    //  [self encodeDouble:[value doubleValue]];
  }



  void _encodeDate(value) {
    throw "Error";
    //  [self encodeUnsignedChar:kAMF0DateType];
    //  [self encodeDouble:([value timeIntervalSince1970] * 1000)];
    //  [self encodeUnsignedShort:([[NSTimeZone localTimeZone] secondsFromGMT] / 60)];
  }

  void _encodeNull() {
    encodeUnsignedChar(AMF0Type.kAMF0NullType);
  }

  void _encodeXML(value) {
    encodeUnsignedChar(AMF0Type.kAMF0XMLObjectType);

  }
}

class AMF3Archiver extends AMFArchiver {
  List _stringTable = [];
  List _traitsTable = [];

  AMF3Archiver.create(): super.create() {}

  void encodeUTF(String s) {
    if (s == null) {
      encodeUnsignedInt29((0 << 1) | 1);
      return;
    }
    List codeUnits = s.codeUnits;
    encodeUnsignedInt29(codeUnits.length);
    encodeDataObject(codeUnits);
  }

  void encodeBool(bool value) {
    encodeUnsignedChar(value ? AMF3Type.kAMF3TrueType : AMF3Type.kAMF3FalseType
        );
  }

  void encodeObject(o) {
    /* if( o is List ) {
      // Todo: Figure this out.
      FlexArrayCollection fac = new FlexArrayCollection();
      fac.source = o;
      _encodeCustomObject(o);
      return;
    }*/

    super.encodeObject(o);
  }


  void _encodeCustomObject(value) {
    if (value is Int8List) _encodeData(value); else super._encodeCustomObject(
        value);

  }

  void _encodeArray(List value) {
    encodeUnsignedChar(AMF3Type.kAMF3ArrayType);
    if (_objectTable.indexOf(value) >= 0) {
      encodeUnsignedInt29(_objectTable.indexOf(value) << 1);
      return;
    }

    _objectTable.add(value);
    encodeUnsignedInt29((value.length << 1) | 1);
    encodeUnsignedChar((0 << 1) | 1);
    value.forEach((e) {
      encodeObject(e);
    });


  }

  void _encodeString(String value, bool omitType) {
    if (!omitType) {
      encodeUnsignedChar(AMF3Type.kAMF3StringType);

    }
    if (value == null || value.length == 0) {
      encodeUnsignedChar((0 << 1) | 1);
      return;
    }
    if (_stringTable.contains(value)) {
      encodeUnsignedInt29(_stringTable.indexOf(value) << 1);
      return;
    }
    _stringTable.add(value);


    List codeUnits = value.codeUnits;

    encodeUnsignedInt29(((codeUnits.length) << 1) | 1);

    encodeDataObject(codeUnits);

  }

  void _encodeDictionary(value) {
    throw "Error";
    //  for (id key in value)
    //  {
    //    if ([key isKindOfClass:[NSNumber class]])
    //    {
    //      [self _encodeMixedArray:value];
    //      return;
    //    }
    //  }
    //  [self _encodeASObject:[ASObject asObjectWithDictionary:value]];
  }

  void _encodeMixedArray(value) {
    throw "Error";
    //  [self encodeUnsignedChar:kAMF3ArrayType];
    //  if ([m_objectTable indexOfObjectIdenticalTo:value] != NSNotFound)
    //  {
    //    [self encodeUnsignedInt29([m_objectTable indexOfObject:value] << 1)];
    //    return;
    //  }
    //
    //  NSMutableArray *numericKeys = [[[NSMutableArray alloc] init] autorelease];
    //  NSMutableArray *stringKeys = [[[NSMutableArray alloc] init] autorelease];
    //  for (id key in value)
    //  {
    //    if ([key isKindOfClass:[NSString class]])
    //      [stringKeys addObject:key];
    //    else if ([key isKindOfClass:[NSNumber class]])
    //      [numericKeys addObject:key];
    //    else
    //      [NSException raise:NSInternalInconsistencyException
    //        format:@"Cannot encode dictionary with key of class %@", [key className]];
    //  }
    //  [self encodeUnsignedInt29(([numericKeys count] << 1) | 1)];
    //  for (NSString *key in stringKeys)
    //  {
    //    [self _encodeString:key omitType:YES];
    //    [self encodeObject:[value objectForKey:key]];
    //  }
    //  [self encodeUnsignedChar((0 << 1) | 1)];
    //  for (NSNumber *key in numericKeys)
    //  {
    //    [self encodeObject:[value objectForKey:key]];
    //  }
  }

  void _encodeDate(value) {
    throw "Error";
    //  [self encodeUnsignedChar:kAMF3DateType];
    //  if ([m_objectTable indexOfObjectIdenticalTo:value] != NSNotFound)
    //  {
    //    [self encodeUnsignedInt29([m_objectTable indexOfObject:value] << 1)];
    //    return;
    //  }
    //  [m_objectTable addObject:value];
    //  [self encodeUnsignedInt29((0 << 1) | 1)];
    //  [self encodeDouble([value timeIntervalSince1970] * 1000)];
  }

  void _encodeData(value) {
    throw "Error";
    //  if ([m_objectTable indexOfObjectIdenticalTo:value] != NSNotFound)
    //  {
    //    [self encodeUnsignedInt29([m_objectTable indexOfObject:value] << 1)];
    //    return;
    //  }
    //  [m_objectTable addObject:value];
    //  [self encodeDataObject:value];
  }

  void _encodeXML(value) {
    encodeUnsignedChar(AMF3Type.kAMF3XMLType);
    _encodeString(value, true);
  }

  void _encodeNumber(value) {
    throw "Error";
    //  if ([[value className] isEqualToString:@"NSCFBoolean"])
    //  {
    //    [self encodeUnsignedChar([value boolValue] ? kAMF3TrueType : kAMF3FalseType)];
    //    return;
    //  }
    //  if (strcmp([value objCType], "f") == 0 ||
    //    strcmp([value objCType], "d") == 0)
    //  {
    //    [self encodeUnsignedChar:kAMF3DoubleType];
    //    [self encodeDouble:[value doubleValue]];
    //    return;
    //  }
    //  [self encodeUnsignedChar:kAMF3IntegerType];
    //  [self encodeUnsignedInt29:[value intValue]];
  }
  
  void _encodeBool(bool value) {
      encodeUnsignedChar(value ? AMF3Type.kAMF3TrueType : AMF3Type.kAMF3FalseType);      
   }

  void _encodeASObject(ASObject value) {

    encodeUnsignedChar(AMF3Type.kAMF3ObjectType);
    if (_objectTable.indexOf(value) != -1) {
      encodeUnsignedInt29((_objectTable.indexOf(value)) << 1);
      return;
    }
    _objectTable.add(value);

    AMF3TraitsInfo traits = new AMF3TraitsInfo();
    traits._externalizable = value._isExternalizable;
    traits._dynamic = (value._type == null || value._type.length == 0);
    traits._count = (traits._dynamic || traits._externalizable ? 0 :
        value.count);
    traits._className = value._type;
    traits._properties = (traits._dynamic ? null : value._properties.keys);
    _encodeTraits(traits);

    if (value._isExternalizable) {
      value._data.forEach((e) {
        encodeObject(e);
      });
    }

    value._properties.forEach((k, v) {
      if (traits._dynamic) {
        // ?
        _encodeString(k.toString(), true);
      }
      // TODO :More guff here
      encodeObject(v);
    });

    if (traits._dynamic) {
      encodeUnsignedInt29((0 << 1) | 1);
    }
  }

  void _encodeTraits(AMF3TraitsInfo traits) {
    if (_traitsTable.contains(traits)) {
      encodeUnsignedInt29((_traitsTable.indexOf(traits) << 2) | 1);
      return;
    }
    _traitsTable.add(traits);
    var infoBits = 3;
    if (traits._externalizable) infoBits |= 4;
    if (traits._dynamic) infoBits |= 8;
    infoBits |= (traits._count << 4);
    encodeUnsignedInt29(infoBits);
    _encodeString(traits._className, true);

    for (String s in traits._properties) {
      //String s = traits._properties[i];
      _encodeString(s, true);
    }
  }

  void _encodeNull() {
    encodeUnsignedChar(AMF3Type.kAMF3NullType);
  }

  void _encodeDouble(double value) {
    encodeUnsignedChar(AMF3Type.kAMF3DoubleType);
    encodeDouble(value);
  }

  void _encodeInteger(int value) {
    encodeUnsignedChar(AMF3Type.kAMF3IntegerType);
    encodeDouble(value.toDouble());
  }
}
