part of amf;

class AMFUnarchiver {
  ByteData m_data;
  int m_position = 0;
  List _objectTable = [];

  ASObject _currentDeserializedObject;


  factory AMFUnarchiver(int amfVersion, ByteData data) {
    if (amfVersion == 3) return new AMF3Unarchiver.create(data); else return
        new AMF0Unarchiver.create(data);
  }

  AMFUnarchiver.create(ByteData data) {
    m_data = data;
  }



  _ensureLength(int len) {}


  int decodeUnsignedShort() {
    _ensureLength(2);
    int value = m_data.getUint16(m_position);
    m_position += 2;
    return value;
  }

  bool decodeBool() {
    return decodeUnsignedChar() != 0;
  }

  int decodeChar() {
    _ensureLength(1);
    int value = m_data.getInt8(m_position);
    m_position++;
    return value;
  }

  int decodeUnsignedChar() {
    return decodeChar();
  }

  int decodeUnsignedInt() {
    _ensureLength(4);
    int value = m_data.getUint32(m_position);
    m_position += 4;
    return value;
  }

  double decodeDouble() {
    _ensureLength(8);
    double value = m_data.getFloat64(m_position);
    m_position += 8;
    return value;
  }

  Object decodeObject() {
    throw "Abstract";
  }

  Object decodeObjectForKey(String key) {
    return _currentDeserializedObject._properties[key];
  }

  double decodeDoubleForKey(String key) {
    return _currentDeserializedObject._properties[key];
  }

  String decodeUTF() {
    return decodeUTFBytes(decodeUnsignedShort());
  }

  String decodeUTFBytes(int length) {
    if (length == 0) return "";
    _ensureLength(length);

    String result = "";

    for (int i = 0; i < length; i++) {
      int char = m_data.getUint8(m_position++);
      result += new String.fromCharCode(char);
    }
    return result;
  }

  int decodeUnsignedInt29() {
    int value;
    int ch = decodeUnsignedChar() & 0xFF;

    if (ch < 128) {
      return ch;
    }

    value = (ch & 0x7F) << 7;
    ch = decodeUnsignedChar() & 0xFF;
    if (ch < 128) {
      return value | ch;
    }

    value = (value | (ch & 0x7F)) << 7;
    ch = decodeUnsignedChar() & 0xFF;
    if (ch < 128) {
      return value | ch;
    }

    value = (value | (ch & 0x7F)) << 8;
    ch = decodeUnsignedChar() & 0xFF;
    return value | ch;
  }

  Object _deserializeObject(ASObject object) {
    if (object._type == null) return object;
    String classname = object._type;

    // Options around FlexArraycollection

    ClassMirror cm = AMF.findClassMirror(classname);
    if (cm == null) {
      if (object._isExternalizable) throw new ArgumentError(
          "No class mirror for ${classname} but it should be externalizable"); else return
          object; // Just use the AS representation.
    }

    ASObject lastDeserializedObject = _currentDeserializedObject;
    _currentDeserializedObject = object;

    Object desObject = AMF.createClassByName(classname);

    _objectTable[_objectTable.indexOf(object)] = desObject;

    if (desObject is Externalizable) desObject.readObject(this); else {
      // Party like it's 1599
      print("FISH");
      InstanceMirror instanceMirror = reflect(desObject);
      for (String key in object._properties.keys) {
        print("${key}");

        instanceMirror.setField(new Symbol(key), object._properties[key]);
      }
    }
    _currentDeserializedObject = lastDeserializedObject;
    return desObject;
  }


}

class AMF0Unarchiver extends AMFUnarchiver {
  AMF0Unarchiver.create(ByteData data): super.create(data) {

  }

  Object decodeObject() {
    int type = decodeUnsignedChar(); // AMF0Type
    return _decodeObjectWithType(type);
  }

  Object _decodeObjectWithType(int type) {

    Object value = null;
    switch (type) {
      case AMF0Type.kAMF0NumberType:
        //value = decodeDouble();
        break;

      case AMF0Type.kAMF0BooleanType:
        value = decodeBool();
        break;

      case AMF0Type.kAMF0StringType:
        value = decodeUTF();
        break;

      case AMF0Type.kAMF0AVMPlusObjectType:
        {
          ByteBuffer bb = m_data.buffer;
          var dataRange = new ByteData.view(bb, m_position);

          AMFUnarchiver amf3Unarchiver = new AMFUnarchiver(3, dataRange);
          //[[AMFUnarchiver alloc]
          //  initForReadingWithData:[m_data subdataWithRange:
          //   (NSRange){m_position, [m_data length] - m_position}]
          //  encoding:kAMF3Version];
          value = amf3Unarchiver.decodeObject();
          //m_position += [m_data length] - m_position - [amf3Unarchiver bytesAvailable];

        }
        break;
      case AMF0Type.kAMF0StrictArrayType:
        value = _decodeArray();
        break;

      case AMF0Type.kAMF0TypedObjectType:
        value = _decodeTypedObject();
        break;

      case AMF0Type.kAMF0LongStringType:
        value = _decodeLongString();
        break;

      case AMF0Type.kAMF0ObjectType:
        value = _decodeASObject(null);
        break;

      case AMF0Type.kAMF0XMLObjectType:
        value = _decodeXML();
        break;

      case AMF0Type.kAMF0NullType:
        value = null;
        break;

      case AMF0Type.kAMF0DateType:
        value = _decodeDate();
        break;

      case AMF0Type.kAMF0ECMAArrayType:
        value = _decodeECMAArray();
        break;

      case AMF0Type.kAMF0ReferenceType:
        value = _decodeReference();
        break;

      case AMF0Type.kAMF0UndefinedType:
        value = null;
        break;

      case AMF0Type.kAMF0UnsupportedType:
        throw new StateError("Unsupported type");

      case AMF0Type.kAMF0ObjectEndType:
        throw new StateError("Unexpected object end");

      case AMF0Type.kAMF0RecordsetType:
        throw new StateError("Unexpected recordset");

      default:
        throw new StateError("Unknown type");
    }

    return value;
  }

  String _decodeLongString() {
    int length = decodeUnsignedInt();
    if (length == 0) return "";
    return decodeUTFBytes(length);
  }

  Object _decodeXML() {
    String xmlString = _decodeLongString();
    return xmlString;
  }

}

class AMF3Unarchiver extends AMFUnarchiver implements ObjectInput {
  List _stringTable = new List();
  List _traitsTable = new List();

  AMF3Unarchiver.create(ByteData data): super.create(data) {

  }

  // ObjectInput---:
  Object readObject() {
    return decodeObject();
  }

  String readUTF() {
    return decodeUTFBytes(decodeUnsignedShort());
  }

  int readInt() {
    return decodeUnsignedInt();
  }

  bool readBool() {
    return decodeBool();
  }

  // --- ObjectInput
  Object decodeObject() {
    int type = decodeUnsignedChar(); // AMF0Type
    return _decodeObjectWithType(type);
  }


  Object _decodeObjectWithType(int type) {
    Object value = null;
    switch (type) {
      case AMF3Type.kAMF3StringType:
        value = decodeUTF();
        break;

      case AMF3Type.kAMF3ObjectType:
        value = _decodeASObject();
        break;

      case AMF3Type.kAMF3ArrayType:
        value = _decodeArray();
        break;

      case AMF3Type.kAMF3FalseType:
        value = false;
        break;

      case AMF3Type.kAMF3TrueType:
        value = true;
        break;

      case AMF3Type.kAMF3IntegerType:
        {
          int intValue = decodeUnsignedInt29();
          intValue = (intValue << 3) >> 3;
          value = intValue;
          break;
        }

      case AMF3Type.kAMF3DoubleType:
        value = decodeDouble();
        break;

      case AMF3Type.kAMF3UndefinedType:
        return null;


      case AMF3Type.kAMF3NullType:
        return null;

      case AMF3Type.kAMF3XMLType:
      case AMF3Type.kAMF3XMLDocType:
        value = _decodeXML();
        break;

      case AMF3Type.kAMF3DateType:
        value = _decodeDate();
        break;

      case AMF3Type.kAMF3ByteArrayType:
        value = _decodeByteArray();
        break;

      case AMF3Type.kDictionaryType:
        throw new StateError("Don't support dictionary type");


      default:
        throw new StateError("Unknown type");

    }

    return value;
  }

  Object _decodeASObject() {
    int ref = decodeUnsignedInt29();
    if ((ref & 1) == 0) {
      ref = (ref >> 1);
      return _objectReferenceAtIndex(ref);
    }

    AMF3TraitsInfo traitsInfo = _decodeTraits(ref);
    Object object;
    if (traitsInfo._className != null && traitsInfo._className.length > 0) {
      object = new ASObject();
      object._type = traitsInfo._className;
      object._isExternalizable = traitsInfo._externalizable;
    } else {
      object = new Map();
    }

    _objectTable.add(object);


    String key;
    for (key in traitsInfo._properties) {
      object[key] = decodeObject();
    }

    if (traitsInfo._dynamic) {
      key = decodeUTF();
      while (key != null && key.length > 0) {
        object[key] = decodeObject();
        key = decodeUTF();
      }
    }

    if (!(object is ASObject)) {
      return object;
    }

    Object desObject = _deserializeObject(object);
    if (desObject == object) {
      return object;
    }
    //[m_objectTable replaceObjectAtIndex:[m_objectTable indexOfObject:object] withObject:desObject];

    return desObject;
  }

  Object _decodeArray() {
    int ref = decodeUnsignedInt29();
    if ((ref & 1) == 0) {
      ref = (ref >> 1);
      return _objectReferenceAtIndex(ref);
    }
    int length = (ref >> 1);
    Object array;
    for ( ; ; ) {
      String name = decodeUTF();
      if (name == null || name.length == 0) break;
      if (array == null) {
        array = new Map();
        _objectTable.add(array);
      }
      array[name] = decodeObject();
    }

    if (array == null) {
      array = [];
      _objectTable.add(array);
      for (int i = 0; i < length; i++) {
        array.add(decodeObject());
      }
    } else {
      for (int i = 0; i < length; i++) {
        array[i] = decodeObject();
      }
    }
    return array;
  }

  Object _decodeXML() {
    String xmlString = decodeUTF();
    return new XML(xmlString);
  }


  Object _objectReferenceAtIndex(int idx) {
    return _objectTable[idx];
  }

  AMF3TraitsInfo _decodeTraits(int infoBits) {
    if ((infoBits & 3) == 1) {
      infoBits = (infoBits >> 2);
      return _traitsReferenceAtIndex(infoBits);
    }
    bool externalizable = (infoBits & 4) == 4;
    bool dynamic = (infoBits & 8) == 8;
    int count = infoBits >> 4;
    String className = decodeUTF();

    AMF3TraitsInfo info = new AMF3TraitsInfo();
    info._className = className;
    info._dynamic = dynamic;
    info._externalizable = externalizable;
    info._count = count;
    while ((count--) > 0) // TODO: maybe >= 0..
    {
      info.addProperty(decodeUTF());
    }

    _traitsTable.add(info);

    return info;
  }

  String decodeUTF() {
    int ref = decodeUnsignedInt29();
    if ((ref & 1) == 0) {
      ref = (ref >> 1);
      return _stringReferenceAtIndex(ref);
    }
    int length = ref >> 1;
    if (length == 0) {
      return "";
    }
    String value = decodeUTFBytes(length);
    _stringTable.add(value);
    return value;
  }

  String _stringReferenceAtIndex(int index) {
    return _stringTable[index];
  }

  AMF3TraitsInfo _traitsReferenceAtIndex(int index) {
    return _traitsTable[index];
  }

}
