part of amf;

class FlexCommandMessageOperationType {
  static const int kFlexCommandMessageSubscribeOperation = 0;

  static const int kFlexCommandMessageUnsubscribeOperation = 1;

  static const int kFlexCommandMessagePollOperation = 2;

  static const int kFlexCommandMessageClientSyncOperation = 4;

  static const int kFlexCommandMessageClientPingOperation = 5;

  static const int kFlexCommandMessageClusterRequestOperation = 7;

  static const int kFlexCommandMessageLoginOperation = 8;

  static const int kFlexCommandMessageLogoutOperation = 9;

  static const int kFlexCommandMessageSubscriptionInvalidateOperation = 10;

  static const int kFlexCommandMessageMultiSubscribeOperation = 11;

  static const int kFlexCommandMessageDisconnectOperation = 12;

  static const int kFlexCommandMessageTriggerConnectOperation = 13;

  static const int kFlexCommandMessageUnknownOperation = 1000;
}

class RemoteClass {
  final String name;
  const RemoteClass([this.name]);
}

@RemoteClass("flex.messaging.io.ArrayCollection")
class FlexArrayCollection implements Externalizable {
  List source;

  void writeObject(ObjectOutput out) {
    out.writeObject(source);
  }
  void readObject(ObjectInput input) {
    source = input.readObject();
  }
}

class FlexObjectProxy {
  Object object;
}

class FlexAbstractMessage implements Externalizable {
  Object body;

  String clientId;

  String destination;

  var headers;

  // dict

  String messageId;

  int timeToLive = 0;

  int timestamp = 0;

  FlexAbstractMessage() {
    messageId = "4D8845C0-6768-4942-9F9E-FC45E0215C8B";
    clientId = "4D8845C0-6768-4942-9F9E-FC45E021FFFF";
    timestamp = new DateTime.now().millisecondsSinceEpoch;
  }

  void writeObject(var coder) {
    coder.encodeObjectKeyValue("body", body);
    coder.encodeObjectKeyValue("clientId", clientId);
    coder.encodeObjectKeyValue("destination", destination);
    coder.encodeObjectKeyValue("headers", headers);
    coder.encodeObjectKeyValue("messageId", messageId);
    coder.encodeDoubleKeyValue("timeToLive", (timeToLive * 1000).toDouble());
    coder.encodeDoubleKeyValue("timestamp", (timestamp * 1000).toDouble());
  }

  void readObject(var coder) {
    body = coder.decodeObjectForKey("body");
    clientId = coder.decodeObjectForKey("clientId");
    destination = coder.decodeObjectForKey("destination");
    headers = coder.decodeObjectForKey("headers");
    messageId = coder.decodeObjectForKey("messageId");
    timeToLive = (coder.decodeDoubleForKey("timeToLive") / 1000).toInt();
    timestamp = (coder.decodeDoubleForKey("timestamp") / 1000).toInt();

  }
}

class FlexAsyncMessage extends FlexAbstractMessage {
  String correlationId;

  void writeObject(var coder) {
    super.writeObject(coder);
    coder.encodeObjectKeyValue("correlationId", correlationId);
  }
}

class FlexCommandMessage extends FlexAsyncMessage {
  int operation;

  void writeObject(var coder) {
    super.writeObject(coder);
    coder.encodeObjectKeyValue("operation", operation);
  }
}

@RemoteClass("flex.messaging.messages.AcknowledgeMessage")
class FlexAcknowledgeMessage extends FlexAsyncMessage {
}

class FlexErrorMessage extends FlexAcknowledgeMessage {
  Object extendedData;

  String faultCode;

  String faultDetail;

  String faultString;

  Object rootCause;

  void writeObject(var coder) {
    super.writeObject(coder);
    coder.encodeObjectKeyValue("extendedData", extendedData);
    coder.encodeObjectKeyValue("faultCode", faultCode);
    coder.encodeObjectKeyValue("faultDetail", faultDetail);
    coder.encodeObjectKeyValue("faultString", faultString);
    coder.encodeObjectKeyValue("rootCause", rootCause);
  }

}

@RemoteClass("flex.messaging.messages.RemotingMessage")
class FlexRemotingMessage extends FlexAbstractMessage {
  String operation;

  String source;

  void writeObject(var coder) {
    super.writeObject(coder);
    coder.encodeObjectKeyValue("operation", operation);
    coder.encodeObjectKeyValue("source", source);
  }
}
