part of amf;

class GDSAsyncToken {
  FlexAbstractMessage _message;
  List _responders;
  GDSAbstractEvent _responseEvent;

  GDSAsyncToken(FlexAbstractMessage message) {
    _message = message;
    _responders = [];
  }
}

class GDSChannel {
  String _id;
  String _url;
  String _credentials;

  bool _connected;
  bool _authenticated;

  int _timeout;

  List _pendingTokens;
  List _sentTokens;

  GDSXAMFRequest _request;

  GDSChannel(String id, String url) {
    _id = id;
    _url = url;
  }

  Future sendToken(GDSAsyncToken token) {
    // Some kind of parallel stuff

    List messages = [];

    FlexAbstractMessage message = token._message;

    messages.add(message);

    AMFActionMessage amf0 = _createAMF0MessageWithMessages(messages);

    var data = amf0.data;

    return _atomicSendRequest(data);

  }

  Future _atomicSendRequest(var data) {
    _request = new GDSXAMFRequest();
    return _request.process(_url, data);
  }

  AMFActionMessage _createAMF0MessageWithMessages(List messages) {
    AMFActionMessage amf0 = new AMFActionMessage();
    messages.forEach((E) {
      amf0.addBody("", "/1", [E]);
    });
    return amf0;
  }
}

class GDSChannelNotification {
  GDSChannel _channel;
  int _level;
  String _message;
  var _error;
  var _exception;

  Map _userInfo;
}


class GDSConfiguration {

}

class GDSAbstractEvent {

}

class GDSFaultEvent {
  FlexErrorMessage _response;
}


class GDSResultEvent extends GDSAbstractEvent {
  FlexAcknowledgeMessage _response;

  GDSResultEvent(this._response);

  Object get result => _response.body;
}


/*
class GDSPersistentList extends FlexArrayCollection {
  int      _initialized;
  String   _metadata;
  int      _dirty;
}

class GDSPersistentSet extends GDSPersistentList {

}
*/
class GDSRemoteObject {
  GDSChannel _channel;
  String _destination;

  GDSRemoteObject(GDSChannel channel, String destination) {
    _channel = channel;
    _destination = destination;
  }

  Future callMethod(String method, List args) {
    FlexRemotingMessage message = new FlexRemotingMessage();
    message.body = (args == null ? [] : args);
    message.operation = method;
    message.destination = _destination;

    GDSAsyncToken token = new GDSAsyncToken(message);
    /*
     * Options to call method with responders
     */

    return _channel.sendToken(token);
  }
}

class GDSRemoteClass {}


class GDSUtil {}

class GDSXAMFRequest {
  Future process(String url, var body) {

    HttpRequest request = new HttpRequest();

    request.open('POST', url);
    request.responseType = "arraybuffer";
    request.setRequestHeader("Content-Type", "application/x-amf");
    //request.setRequestHeader("User-Agent", "DartAMF");

    return HttpRequest.request(url, method: "POST", responseType: "arraybuffer",
        sendData: body).then((request) {
      //this is a hack for dart2js because of a bug
      Uint8List buffer = new Uint8List.view(request.response, 0,
          (request.response as ByteBuffer).lengthInBytes);

      return decodeResponse(buffer);
    });

  }

  Object decodeResponse(Uint8List data) {
    AMFActionMessage message = new AMFActionMessage.initWithData(data);

    List responses = message.bodies;

    int count = responses.length;

    for (int i = 0; i < count; i++) {
      Object object = responses[i].data;

      if (object is FlexAsyncMessage) {
        //FlexAsyncMessage response = (FlexAsyncMessage)object;
        GDSAbstractEvent event;

        if (object is FlexErrorMessage) throw object; else if (object is
            FlexAcknowledgeMessage) {
          event = new GDSResultEvent(object);
          return event;
        }

      }

    }
    return null;
  }
}
